from functools import partial
import logging
from unittest import mock, TestCase

from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs.common.reports import ReportItemSeverity as Severities
from pcs.common.reports import codes as report_codes
from pcs.lib.env import LibraryEnvironment

import pcs.lib.commands.alert as cmd_alert


get_env_tools = partial(
    get_env_tools,
    exception_reports_in_processor_by_default=False,
)


class CreateAlertTest(TestCase):
    fixture_final_alerts = """
        <alerts>
            <alert id="my-alert" path="/my/path" description="my description">
                <meta_attributes id="my-alert-meta_attributes">
                    <nvpair
                        id="my-alert-meta_attributes-meta1"
                        name="meta1"
                        value="val1"
                    />
                </meta_attributes>
                <instance_attributes id="my-alert-instance_attributes">
                    <nvpair
                        id="my-alert-instance_attributes-another"
                        name="another"
                        value="val"
                    />
                    <nvpair
                        id="my-alert-instance_attributes-instance"
                        name="instance"
                        value="value"
                    />
                </instance_attributes>
            </alert>
        </alerts>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_no_path(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.create_alert(
                self.env_assist.get_env(), None, None, None, None
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    {"option_names": ["path"], "option_type": None},
                    None,
                ),
            ],
        )

    def test_create_no_upgrade(self):
        (
            self.config.runner.cib.load().env.push_cib(
                optional_in_conf=self.fixture_final_alerts
            )
        )
        cmd_alert.create_alert(
            self.env_assist.get_env(),
            "my-alert",
            "/my/path",
            {"instance": "value", "another": "val"},
            {"meta1": "val1"},
            "my description",
        )


class UpdateAlertTest(TestCase):
    fixture_initial_alerts = """
        <alerts>
            <alert id="my-alert" path="/my/path" description="my description">
                <instance_attributes id="my-alert-instance_attributes">
                    <nvpair
                        id="my-alert-instance_attributes-instance"
                        name="instance"
                        value="value"
                    />
                    <nvpair
                        id="my-alert-instance_attributes-another"
                        name="another"
                        value="val"
                    />
                </instance_attributes>
                <meta_attributes id="my-alert-meta_attributes">
                    <nvpair
                        id="my-alert-meta_attributes-meta1"
                        name="meta1"
                        value="val1"
                    />
                </meta_attributes>
            </alert>
        </alerts>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_update_all(self):
        fixture_final_alerts = """
        <alerts>
            <alert id="my-alert" path="/another/one">
                <instance_attributes id="my-alert-instance_attributes">
                    <nvpair
                        id="my-alert-instance_attributes-another"
                        name="another"
                        value="val"
                    />
                    <nvpair
                        id="my-alert-instance_attributes-my-attr"
                        name="my-attr"
                        value="its_val"
                    />
                </instance_attributes>
                <meta_attributes id="my-alert-meta_attributes">
                    <nvpair
                        id="my-alert-meta_attributes-meta1"
                        name="meta1"
                        value="val2"
                    />
                </meta_attributes>
            </alert>
        </alerts>
        """
        (
            self.config.runner.cib.load(
                optional_in_conf=self.fixture_initial_alerts
            ).env.push_cib(
                replace={"./configuration/alerts": fixture_final_alerts}
            )
        )
        cmd_alert.update_alert(
            self.env_assist.get_env(),
            "my-alert",
            "/another/one",
            {"instance": "", "my-attr": "its_val"},
            {"meta1": "val2"},
            "",
        )

    def test_update_instance_attribute(self):
        (
            self.config.runner.cib.load(
                optional_in_conf=self.fixture_initial_alerts
            ).env.push_cib(
                replace={
                    './configuration/alerts/alert[@id="my-alert"]/'
                    'instance_attributes/nvpair[@name="instance"]': """
                        <nvpair
                            id="my-alert-instance_attributes-instance"
                            name="instance"
                            value="new_val"
                        />
                    """
                }
            )
        )
        cmd_alert.update_alert(
            self.env_assist.get_env(),
            "my-alert",
            None,
            {"instance": "new_val"},
            {},
            None,
        )

    def test_alert_doesnt_exist(self):
        (
            self.config.runner.cib.load(
                optional_in_conf="""
                    <alerts>
                        <alert id="alert" path="path"/>
                    </alerts>
                """
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.update_alert(
                self.env_assist.get_env(), "unknown", "test", {}, {}, None
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "context_type": "alerts",
                        "context_id": "",
                        "id": "unknown",
                        "expected_types": ["alert"],
                    },
                    None,
                ),
            ],
        )


class RemoveAlertTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            optional_in_conf="""
                <alerts>
                    <alert id="alert1" path="path"/>
                    <alert id="alert2" path="/path"/>
                    <alert id="alert3" path="/path"/>
                    <alert id="alert4" path="/path"/>
                </alerts>
            """
        )

    def test_one_alert(self):
        self.config.env.push_cib(
            remove="./configuration/alerts/alert[@id='alert2']"
        )
        cmd_alert.remove_alert(self.env_assist.get_env(), ["alert2"])

    def test_multiple_alerts(self):
        self.config.env.push_cib(
            remove=[
                "./configuration/alerts/alert[@id='alert1']",
                "./configuration/alerts/alert[@id='alert3']",
                "./configuration/alerts/alert[@id='alert4']",
            ]
        )
        cmd_alert.remove_alert(
            self.env_assist.get_env(), ["alert1", "alert3", "alert4"]
        )

    def test_no_alert(self):
        self.config.env.push_cib()
        cmd_alert.remove_alert(self.env_assist.get_env(), [])

    def test_alerts_dont_exist(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.remove_alert(
                self.env_assist.get_env(),
                ["unknown1", "alert1", "unknown2", "alert2"],
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "context_type": "alerts",
                        "context_id": "",
                        "id": "unknown1",
                        "expected_types": ["alert"],
                    },
                    None,
                ),
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "context_type": "alerts",
                        "context_id": "",
                        "id": "unknown2",
                        "expected_types": ["alert"],
                    },
                    None,
                ),
            ]
        )


class AddRecipientTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            optional_in_conf="""
                <alerts>
                    <alert id="alert" path="path">
                        <recipient id="alert-recipient" value="value1"/>
                    </alert>
                </alerts>
            """
        )

    def test_value_not_defined(self):
        self.config.remove("runner.cib.load")
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.add_recipient(
                self.env_assist.get_env(), "unknown", "", {}, {}
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    {"option_names": ["value"], "option_type": None},
                )
            ],
        )

    def test_recipient_already_exists(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.add_recipient(
                self.env_assist.get_env(),
                "alert",
                "value1",
                {},
                {},
                recipient_id="alert-recipient",
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.ID_ALREADY_EXISTS,
                    {"id": "alert-recipient"},
                )
            ],
        )

    def test_without_id(self):
        self.config.env.push_cib(
            replace={
                './/alert[@id="alert"]': """
                <alert id="alert" path="path">
                    <recipient id="alert-recipient" value="value1"/>
                    <recipient id="alert-recipient-1" value="value">
                        <meta_attributes
                            id="alert-recipient-1-meta_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr2"
                                name="attr2"
                                value="val2"
                            />
                        </meta_attributes>
                        <instance_attributes
                            id="alert-recipient-1-instance_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-instance_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                        </instance_attributes>
                    </recipient>
                </alert>
                """
            }
        )
        cmd_alert.add_recipient(
            self.env_assist.get_env(),
            "alert",
            "value",
            {"attr1": "val1"},
            {"attr2": "val2", "attr1": "val1"},
        )

    def test_with_id(self):
        self.config.env.push_cib(
            replace={
                './/alert[@id="alert"]': """
                <alert id="alert" path="path">
                    <recipient id="alert-recipient" value="value1"/>
                    <recipient id="my-recipient" value="value">
                        <meta_attributes
                            id="my-recipient-meta_attributes"
                        >
                            <nvpair
                                id="my-recipient-meta_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                            <nvpair
                                id="my-recipient-meta_attributes-attr2"
                                name="attr2"
                                value="val2"
                            />
                        </meta_attributes>
                        <instance_attributes
                            id="my-recipient-instance_attributes"
                        >
                            <nvpair
                                id="my-recipient-instance_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                        </instance_attributes>
                    </recipient>
                </alert>
                """
            }
        )
        cmd_alert.add_recipient(
            self.env_assist.get_env(),
            "alert",
            "value",
            {"attr1": "val1"},
            {"attr2": "val2", "attr1": "val1"},
            recipient_id="my-recipient",
        )


class UpdateRecipientTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            optional_in_conf="""
                <alerts>
                    <alert id="alert" path="path">
                        <recipient id="alert-recipient" value="value1"/>
                        <recipient id="alert-recipient-1" value="value"
                            description="d"
                        >
                            <meta_attributes
                                id="alert-recipient-1-meta_attributes"
                            >
                                <nvpair
                                    id="alert-recipient-1-meta_attributes-attr1"
                                    name="attr1"
                                    value="val1"
                                />
                                <nvpair
                                    id="alert-recipient-1-meta_attributes-attr2"
                                    name="attr2"
                                    value="val2"
                                />
                            </meta_attributes>
                            <instance_attributes
                                id="alert-recipient-1-instance_attributes"
                            >
                                <nvpair
                                    id="alert-recipient-1-instance_attributes-attr1"
                                    name="attr1"
                                    value="val1"
                                />
                            </instance_attributes>
                        </recipient>
                    </alert>
                </alerts>
            """
        )

    def test_empty_value(self):
        self.config.remove("runner.cib.load")
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.update_recipient(
                self.env_assist.get_env(),
                "alert-recipient-1",
                {},
                {},
                recipient_value="",
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.CIB_ALERT_RECIPIENT_VALUE_INVALID,
                    {"recipient": ""},
                )
            ],
        )

    def test_recipient_not_found(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.update_recipient(
                self.env_assist.get_env(), "recipient", {}, {}
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "id": "recipient",
                        "expected_types": ["recipient"],
                        "context_id": "",
                        "context_type": "alerts",
                    },
                    None,
                )
            ],
        )

    def test_update_all(self):
        self.config.env.push_cib(
            replace={
                './/alert[@id="alert"]': """
                <alert id="alert" path="path">
                    <recipient id="alert-recipient" value="value1"/>
                    <recipient
                        id="alert-recipient-1"
                        value="new_val"
                        description="desc"
                    >
                        <meta_attributes
                            id="alert-recipient-1-meta_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr2"
                                name="attr2"
                                value="val2"
                            />
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr3"
                                name="attr3"
                                value="new_val"
                            />
                        </meta_attributes>
                        <instance_attributes
                            id="alert-recipient-1-instance_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-instance_attributes-attr1"
                                name="attr1"
                                value="value"
                            />
                        </instance_attributes>
                    </recipient>
                </alert>
                """,
            }
        )
        cmd_alert.update_recipient(
            self.env_assist.get_env(),
            "alert-recipient-1",
            {"attr1": "value"},
            {"attr1": "", "attr3": "new_val"},
            recipient_value="new_val",
            description="desc",
        )


class RemoveRecipientTest(TestCase):
    fixture_initial_alerts = """
        <alerts>
            <alert id="alert" path="path">
                <recipient id="alert-recip1" value="value1"/>
                <recipient id="alert-recip2" value="value2"/>
            </alert>
            <alert id="alert2" path="path">
                <recipient id="alert2-recip3" value="value3"/>
                <recipient id="alert2-recip4" value="value4"/>
            </alert>
        </alerts>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_alerts
        )

    def test_recipient_not_found(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.remove_recipient(
                self.env_assist.get_env(),
                ["recipient", "alert-recip1", "alert2-recip1"],
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "id": "recipient",
                        "expected_types": ["recipient"],
                        "context_type": "alerts",
                        "context_id": "",
                    },
                    None,
                ),
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "id": "alert2-recip1",
                        "expected_types": ["recipient"],
                        "context_type": "alerts",
                        "context_id": "",
                    },
                    None,
                ),
            ]
        )

    def test_one_recipient(self):
        self.config.env.push_cib(
            remove="./configuration/alerts/alert/recipient[@id='alert-recip1']"
        )
        cmd_alert.remove_recipient(self.env_assist.get_env(), ["alert-recip1"])

    def test_multiple_recipients(self):
        self.config.env.push_cib(
            remove=[
                "./configuration/alerts/alert/recipient[@id='alert-recip1']",
                "./configuration/alerts/alert/recipient[@id='alert-recip2']",
                "./configuration/alerts/alert/recipient[@id='alert2-recip4']",
            ]
        )
        cmd_alert.remove_recipient(
            self.env_assist.get_env(),
            ["alert-recip1", "alert-recip2", "alert2-recip4"],
        )

    def test_no_recipient(self):
        self.config.env.push_cib()
        cmd_alert.remove_recipient(self.env_assist.get_env(), [])


@mock.patch.object(LibraryEnvironment, "get_cib")
@mock.patch("pcs.lib.cib.alert.get_all_alerts")
class GetAllAlertsTest(TestCase):
    def setUp(self):
        self.mock_log = mock.MagicMock(spec_set=logging.Logger)
        self.mock_rep = MockLibraryReportProcessor()
        self.mock_env = LibraryEnvironment(self.mock_log, self.mock_rep)

    def test_success(self, mock_alerts, mock_get_cib):
        mock_alerts.return_value = [{"id": "alert"}]
        mock_get_cib.return_value = "<cib/>"
        self.assertEqual(
            [{"id": "alert"}], cmd_alert.get_all_alerts(self.mock_env)
        )
        self.assertEqual(1, mock_alerts.call_count)



from unittest import mock, TestCase

from pcs_test.tools.assertions import ExtendedAssertionsMixin
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

import pcs.lib.commands.acl as cmd_acl
from pcs.lib.env import LibraryEnvironment


class AclCommandsTest(TestCase, ExtendedAssertionsMixin):
    def setUp(self):
        self.mock_rep = MockLibraryReportProcessor()
        self.mock_env = mock.MagicMock(spec_set=LibraryEnvironment)
        self.mock_env.report_processor = self.mock_rep
        self.cib = "cib"
        self.mock_env.get_cib.return_value = self.cib

    def assert_get_cib_called(self):
        self.mock_env.get_cib.assert_called_once_with()

    def assert_same_cib_pushed(self):
        self.mock_env.push_cib.assert_called_once_with()

    def assert_cib_not_pushed(self):
        self.assertEqual(0, self.mock_env.push_cib.call_count)


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
class CibAclSection(TestCase):
    def test_push_cib_on_success(self):
        # pylint: disable=no-self-use
        env = mock.MagicMock()
        env.get_cib = mock.Mock(return_value="cib")
        with cmd_acl.cib_acl_section(env):
            pass
        env.get_cib.assert_called_once_with()
        env.push_cib.assert_called_once_with()

    def test_does_not_push_cib_on_exception(self):
        env = mock.MagicMock()

        def run():
            with cmd_acl.cib_acl_section(env):
                raise AssertionError()

        self.assertRaises(AssertionError, run)
        env.get_cib.assert_called_once_with()
        env.push_cib.assert_not_called()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.validate_permissions")
@mock.patch("pcs.lib.cib.acl.create_role")
@mock.patch("pcs.lib.cib.acl.add_permissions_to_role")
class CreateRoleTest(AclCommandsTest):
    def test_success(self, mock_add_perm, mock_create_role, mock_validate):
        perm_list = ["my", "list"]
        mock_create_role.return_value = "role el"
        cmd_acl.create_role(self.mock_env, "role_id", perm_list, "desc")
        self.assert_get_cib_called()
        mock_validate.assert_called_once_with(self.cib, perm_list)
        mock_create_role.assert_called_once_with(self.cib, "role_id", "desc")
        mock_add_perm.assert_called_once_with("role el", perm_list)
        self.assert_same_cib_pushed()

    def test_no_permission(
        self, mock_add_perm, mock_create_role, mock_validate
    ):
        mock_create_role.return_value = "role el"
        cmd_acl.create_role(self.mock_env, "role_id", [], "desc")
        self.assert_get_cib_called()
        self.assertEqual(0, mock_validate.call_count)
        mock_create_role.assert_called_once_with(self.cib, "role_id", "desc")
        self.assertEqual(0, mock_add_perm.call_count)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.remove_role")
class RemoveRoleTest(AclCommandsTest):
    def test_success(self, mock_remove):
        cmd_acl.remove_role(self.mock_env, "role_id", False)
        self.assert_get_cib_called()
        mock_remove.assert_called_once_with(self.cib, "role_id", False)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.find_target_or_group")
@mock.patch("pcs.lib.cib.acl.assign_role")
class AssignRoleNotSpecific(AclCommandsTest, ExtendedAssertionsMixin):
    def test_success(self, mock_assign, find_target_or_group):
        find_target_or_group.return_value = "target_el"
        cmd_acl.assign_role_not_specific(self.mock_env, "role_id", "target_id")
        self.assert_get_cib_called()
        find_target_or_group.assert_called_once_with(self.cib, "target_id")
        mock_assign.assert_called_once_with(self.cib, "role_id", "target_el")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.find_target")
@mock.patch("pcs.lib.cib.acl.assign_role")
class AssignRoleToTargetTest(AclCommandsTest):
    def test_success(self, mock_assign, find_target):
        find_target.return_value = "target_el"
        cmd_acl.assign_role_to_target(self.mock_env, "role_id", "target_id")
        self.assert_get_cib_called()
        mock_assign.assert_called_once_with(self.cib, "role_id", "target_el")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.find_group")
@mock.patch("pcs.lib.cib.acl.assign_role")
class AssignRoleToGroupTest(AclCommandsTest):
    def test_success(self, mock_assign, find_group):
        find_group.return_value = "group_el"
        cmd_acl.assign_role_to_group(self.mock_env, "role_id", "group_id")
        self.assert_get_cib_called()
        mock_assign.assert_called_once_with(self.cib, "role_id", "group_el")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.unassign_role")
@mock.patch("pcs.lib.cib.acl.find_target_or_group")
class UnassignRoleNotSpecificTest(AclCommandsTest):
    def test_success(self, find_target_or_group, mock_unassign):
        find_target_or_group.return_value = "target_el"
        cmd_acl.unassign_role_not_specific(
            self.mock_env, "role_id", "target_id", False
        )
        self.assert_get_cib_called()
        find_target_or_group.assert_called_once_with(self.cib, "target_id")
        mock_unassign.assert_called_once_with("target_el", "role_id", False)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.unassign_role")
@mock.patch("pcs.lib.cib.acl.find_target")
class UnassignRoleFromTargetTest(AclCommandsTest):
    def test_success(self, find_target, mock_unassign):
        find_target.return_value = "el"
        cmd_acl.unassign_role_from_target(
            self.mock_env, "role_id", "el_id", False
        )
        self.assert_get_cib_called()
        find_target.assert_called_once_with(self.cib, "el_id")
        mock_unassign.assert_called_once_with("el", "role_id", False)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.unassign_role")
@mock.patch("pcs.lib.cib.acl.find_group")
class UnassignRoleFromGroupTest(AclCommandsTest):
    def test_success(self, find_group, mock_unassign):
        find_group.return_value = "el"
        cmd_acl.unassign_role_from_group(
            self.mock_env, "role_id", "el_id", False
        )
        self.assert_get_cib_called()
        find_group.assert_called_once_with(self.cib, "el_id")
        mock_unassign.assert_called_once_with("el", "role_id", False)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.create_target")
@mock.patch("pcs.lib.cib.acl.assign_all_roles")
class CreateTargetTest(AclCommandsTest):
    def test_success(self, mock_assign, mock_create):
        mock_create.return_value = "el"
        cmd_acl.create_target(
            self.mock_env, "el_id", ["role1", "role2", "role3"]
        )
        self.assert_get_cib_called()
        mock_create.assert_called_once_with(self.cib, "el_id")
        mock_assign(self.cib, "el", ["role1", "role2", "role3"])
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.create_group")
@mock.patch("pcs.lib.cib.acl.assign_all_roles")
class CreateGroupTest(AclCommandsTest):
    def test_success(self, mock_assign, mock_create):
        mock_create.return_value = "el"
        cmd_acl.create_group(
            self.mock_env, "el_id", ["role1", "role2", "role3"]
        )
        self.assert_get_cib_called()
        mock_create.assert_called_once_with(self.cib, "el_id")
        mock_assign(self.cib, "el", ["role1", "role2", "role3"])
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.remove_target")
class RemoveTargetTest(AclCommandsTest):
    def test_success(self, mock_remove):
        cmd_acl.remove_target(self.mock_env, "el_id")
        self.assert_get_cib_called()
        mock_remove.assert_called_once_with(self.cib, "el_id")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.remove_group")
class RemoveGroupTest(AclCommandsTest):
    def test_success(self, mock_remove):
        cmd_acl.remove_group(self.mock_env, "el_id")
        self.assert_get_cib_called()
        mock_remove.assert_called_once_with(self.cib, "el_id")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.validate_permissions")
@mock.patch("pcs.lib.cib.acl.provide_role")
@mock.patch("pcs.lib.cib.acl.add_permissions_to_role")
class AddPermissionTest(AclCommandsTest):
    def test_success(self, mock_add_perm, mock_provide_role, mock_validate):
        mock_provide_role.return_value = "role_el"
        cmd_acl.add_permission(self.mock_env, "role_id", "permission_list")
        self.assert_get_cib_called()
        mock_validate.assert_called_once_with(self.cib, "permission_list")
        mock_provide_role.assert_called_once_with(self.cib, "role_id")
        mock_add_perm.assert_called_once_with("role_el", "permission_list")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.remove_permission")
class RemovePermission(AclCommandsTest):
    def test_success(self, mock_remove):
        cmd_acl.remove_permission(self.mock_env, "id")
        self.assert_get_cib_called()
        mock_remove.assert_called_once_with(self.cib, "id")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.cib.acl.get_target_list")
@mock.patch("pcs.lib.cib.acl.get_group_list")
@mock.patch("pcs.lib.cib.acl.get_role_list")
@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
class GetConfigTest(AclCommandsTest):
    def test_success(self, mock_role, mock_group, mock_target):
        mock_role.return_value = "role"
        mock_group.return_value = "group"
        mock_target.return_value = "target"
        self.assertEqual(
            {
                "target_list": "target",
                "group_list": "group",
                "role_list": "role",
            },
            cmd_acl.get_config(self.mock_env),
        )



import base64
import json

from pcs_test.tools.command_env.mock_node_communicator import (
    place_multinode_call,
)


class CorosyncShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def check_corosync_offline(
        self,
        node_labels=None,
        communication_list=None,
        name="http.corosync.check_corosync_offline",
    ):
        """
        Create a call for checking that corosync is offline

        string name -- the key of this call
        list node_labels -- create success responses from these nodes
        list communication_list -- create custom responses
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/status",
            output='{"corosync":false}',
        )

    def get_corosync_online_targets(
        self,
        node_labels=None,
        communication_list=None,
        name="http.corosync.get_corosync_online_targets",
    ):
        """
        Create a call for getting corosync online targets

        list node_labels -- create success responses from these nodes
        list communication_list -- create custom responses
        string name -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/status",
            output='{"corosync":true}',
        )

    def get_corosync_conf(
        self,
        corosync_conf="",
        node_labels=None,
        communication_list=None,
        name="http.corosync.get_corosync_conf",
    ):
        """
        Create a call for loading corosync.conf text from remote nodes

        string corosync_conf -- corosync.conf text to be loaded
        list node_labels -- create success responses from these nodes
        list communication_list -- create custom responses
        string name -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/get_corosync_conf",
            output=corosync_conf,
        )

    def set_corosync_conf(
        self,
        corosync_conf,
        node_labels=None,
        communication_list=None,
        name="http.corosync.set_corosync_conf",
    ):
        """
        Create a call for sending corosync.conf text

        string corosync_conf -- corosync.conf text to be sent
        list node_labels -- create success responses from these nodes
        list communication_list -- create custom responses
        string name -- the key of this call
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/set_corosync_conf",
            param_list=[("corosync_conf", corosync_conf)],
            output="Succeeded",
        )

    def reload_corosync_conf(
        self,
        node_labels=None,
        communication_list=None,
        name="http.corosync.reload_corosync_conf",
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/reload_corosync_conf",
            output=json.dumps(dict(code="reloaded", message="")),
        )

    def qdevice_client_enable(
        self,
        name="http.corosync.qdevice_client_enable",
        node_labels=None,
        communication_list=None,
    ):
        """
        Create a call for enabling qdevice service

        string name -- the key of this call
        list node_labels -- create success responses from these nodes
        list communication_list -- create custom responses
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/qdevice_client_enable",
            output="corosync-qdevice enabled",
        )

    def qdevice_client_disable(
        self,
        name="http.corosync.qdevice_client_disable",
        node_labels=None,
        communication_list=None,
    ):
        """
        Create a call for disabling qdevice service

        string name -- the key of this call
        list node_labels -- create success responses from these nodes
        list communication_list -- create custom responses
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/qdevice_client_disable",
            output="corosync-qdevice disabled",
        )

    def qdevice_client_start(
        self,
        name="http.corosync.qdevice_client_start",
        node_labels=None,
        communication_list=None,
    ):
        """
        Create a call for starting qdevice service

        string name -- the key of this call
        list node_labels -- create success responses from these nodes
        list communication_list -- create custom responses
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/qdevice_client_start",
            output="corosync-qdevice started",
        )

    def qdevice_client_stop(
        self,
        name="http.corosync.qdevice_client_stop",
        node_labels=None,
        communication_list=None,
    ):
        """
        Create a call for stopping qdevice service

        string name -- the key of this call
        list node_labels -- create success responses from these nodes
        list communication_list -- create custom responses
        """
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/qdevice_client_stop",
            output="corosync-qdevice stopped",
        )

    def qdevice_net_get_ca_cert(
        self,
        ca_cert=b"ca_cert",
        node_labels=None,
        communication_list=None,
        name="http.corosync.qdevice_net_get_ca_cert",
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/qdevice_net_get_ca_certificate",
            output=base64.b64encode(ca_cert),
        )

    def qdevice_net_client_setup(
        self,
        ca_cert=b"ca_cert",
        node_labels=None,
        communication_list=None,
        name="http.corosync.qdevice_net_client_setup",
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/qdevice_net_client_init_certificate_storage",
            param_list=[("ca_certificate", base64.b64encode(ca_cert))],
        )

    def qdevice_net_sign_certificate(
        self,
        cluster_name,
        cert=b"cert",
        signed_cert=b"signed cert",
        node_labels=None,
        communication_list=None,
        name="http.corosync.qdevice_net_sign_certificate",
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/qdevice_net_sign_node_certificate",
            param_list=[
                ("certificate_request", base64.b64encode(cert)),
                ("cluster_name", cluster_name),
            ],
            output=base64.b64encode(signed_cert),
        )

    def qdevice_net_client_import_cert_and_key(
        self,
        cert=b"pk12 cert",
        node_labels=None,
        communication_list=None,
        name="http.corosync.qdevice_net_client_import_cert_and_key",
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/qdevice_net_client_import_certificate",
            param_list=[("certificate", base64.b64encode(cert))],
        )



from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from pcs.common.pacemaker.resource.relations import ResourceRelationType
from pcs.lib.commands import resource


def fixture_primitive(_id, members):
    return dict(
        id=_id,
        type=ResourceRelationType.RSC_PRIMITIVE,
        metadata={
            "id": _id,
            "class": "ocf",
            "provider": "pacemaker",
            "type": "Dummy",
        },
        members=members,
    )


def fixture_primitive_xml(_id):
    return f"""
        <primitive id="{_id}" class="ocf" provider="pacemaker" type="Dummy"/>
    """


def fixture_node(entity, members=None, leaf=False):
    return dict(
        relation_entity=entity,
        is_leaf=leaf,
        members=members or [],
    )


def fixture_order(res1, res2, kind="Mandatory", score=None):
    _id = f"order-{res1}-{res2}"
    out = dict(
        id=_id,
        type=ResourceRelationType.ORDER,
        members=[res1, res2],
        metadata={
            "id": _id,
            "first": res1,
            "first-action": "start",
            "then": res2,
            "then-action": "start",
            "kind": kind,
        },
    )
    if score:
        out["metadata"]["score"] = score
    return out


class GetResourceRelationsTree(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_not_existing_resource(self):
        self.config.runner.cib.load()
        resource_id = "not_existing"
        self.env_assist.assert_raise_library_error(
            lambda: resource.get_resource_relations_tree(
                self.env_assist.get_env(),
                resource_id,
            )
        )
        self.env_assist.assert_reports(
            [fixture.report_not_found(resource_id, context_type="resources")]
        )

    def test_simple(self):
        self.config.runner.cib.load(
            resources="<resources>{}</resources>".format(
                fixture_primitive_xml("d1") + fixture_primitive_xml("d2")
            ),
            constraints="""
            <constraints>
                <rsc_order first="d1" first-action="start"
                    id="order-d1-d2" then="d2" then-action="start"
                    kind="Mandatory"/>
            </constraints>
            """,
        )
        prim_members = ["order-d1-d2"]
        expected = fixture_node(
            fixture_primitive("d1", prim_members),
            [
                fixture_node(
                    fixture_order("d1", "d2"),
                    [fixture_node(fixture_primitive("d2", prim_members))],
                )
            ],
        )
        self.assertEqual(
            expected,
            resource.get_resource_relations_tree(
                self.env_assist.get_env(), "d1"
            ),
        )


class GetResourceRelationsTreeComplex(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            resources="""
                <resources>
                    {primitives}
                    <clone id="c">
                        <group id="cg">
                        {in_group}
                        </group>
                    </clone>
                </resources>
            """.format(
                primitives=(
                    fixture_primitive_xml("d1")
                    + fixture_primitive_xml("d2")
                    + fixture_primitive_xml("d3")
                ),
                in_group=(
                    fixture_primitive_xml("cgd1")
                    + fixture_primitive_xml("cgd2")
                    + fixture_primitive_xml("cgd0")
                ),
            ),
            constraints="""
            <constraints>
              <rsc_order first="d1" first-action="start" id="order-d1-d2"
                  then="d2" then-action="start" kind="Mandatory"/>
              <rsc_order first="cgd1" first-action="start" id="order-cgd1-d2"
                  then="d2" then-action="start" kind="Optional" score="10"/>
              <rsc_order kind="Serialize" symmetrical="true"
                  id="pcs_rsc_order_set_1">
                <resource_set sequential="true" require-all="true"
                    action="start" id="pcs_rsc_set_1">
                  <resource_ref id="d1"/>
                  <resource_ref id="d3"/>
                </resource_set>
                <resource_set action="stop" sequential="false"
                    require-all="false" id="pcs_rsc_set_2">
                  <resource_ref id="cg"/>
                  <resource_ref id="d2"/>
                </resource_set>
              </rsc_order>
            </constraints>
            """,
        )
        self.d1_members = ["order-d1-d2", "pcs_rsc_order_set_1"]
        self.d2_members = [
            "order-d1-d2",
            "order-cgd1-d2",
            "pcs_rsc_order_set_1",
        ]
        self.order_set = dict(
            id="pcs_rsc_order_set_1",
            type=ResourceRelationType.ORDER_SET,
            members=["cg", "d1", "d2", "d3"],
            metadata={
                "id": "pcs_rsc_order_set_1",
                "sets": [
                    {
                        "id": "pcs_rsc_set_1",
                        "metadata": {
                            "id": "pcs_rsc_set_1",
                            "sequential": "true",
                            "require-all": "true",
                            "action": "start",
                        },
                        "members": ["d1", "d3"],
                    },
                    {
                        "id": "pcs_rsc_set_2",
                        "metadata": {
                            "id": "pcs_rsc_set_2",
                            "sequential": "false",
                            "require-all": "false",
                            "action": "stop",
                        },
                        "members": ["cg", "d2"],
                    },
                ],
                "kind": "Serialize",
                "symmetrical": "true",
            },
        )
        self.cg_ent = dict(
            id="cg",
            type=ResourceRelationType.RSC_GROUP,
            members=["pcs_rsc_order_set_1", "inner:cg", "outer:c"],
            metadata=dict(id="cg"),
        )

    def test_d1(self):
        outer_cg = dict(
            id="outer:cg",
            type=ResourceRelationType.OUTER_RESOURCE,
            members=["cg"],
            metadata=dict(id="cg"),
        )
        order_opt = fixture_order("cgd1", "d2", kind="Optional", score="10")
        expected = fixture_node(
            fixture_primitive("d1", self.d1_members),
            [
                fixture_node(
                    fixture_order("d1", "d2"),
                    [
                        fixture_node(
                            fixture_primitive("d2", self.d2_members),
                            [
                                fixture_node(
                                    order_opt,
                                    [
                                        fixture_node(
                                            fixture_primitive(
                                                "cgd1",
                                                ["order-cgd1-d2", "outer:cg"],
                                            ),
                                            [
                                                fixture_node(
                                                    outer_cg,
                                                    [
                                                        fixture_node(
                                                            self.cg_ent,
                                                            leaf=True,
                                                        ),
                                                    ],
                                                )
                                            ],
                                        ),
                                    ],
                                ),
                                fixture_node(self.order_set, leaf=True),
                            ],
                        )
                    ],
                ),
                fixture_node(
                    self.order_set,
                    [
                        fixture_node(
                            self.cg_ent,
                            [
                                fixture_node(
                                    dict(
                                        id="inner:cg",
                                        type=(
                                            ResourceRelationType.INNER_RESOURCES
                                        ),
                                        members=["cgd1", "cgd2", "cgd0"],
                                        metadata=dict(id="cg"),
                                    ),
                                    [
                                        fixture_node(
                                            fixture_primitive(
                                                "cgd1",
                                                ["order-cgd1-d2", "outer:cg"],
                                            ),
                                            leaf=True,
                                        ),
                                        fixture_node(
                                            fixture_primitive(
                                                "cgd2", ["outer:cg"]
                                            ),
                                        ),
                                        fixture_node(
                                            fixture_primitive(
                                                "cgd0", ["outer:cg"]
                                            ),
                                        ),
                                    ],
                                ),
                                fixture_node(
                                    dict(
                                        id="outer:c",
                                        type=(
                                            ResourceRelationType.OUTER_RESOURCE
                                        ),
                                        members=["c"],
                                        metadata=dict(id="c"),
                                    ),
                                    [
                                        fixture_node(
                                            dict(
                                                id="c",
                                                type=ResourceRelationType.RSC_CLONE,
                                                members=["inner:c"],
                                                metadata=dict(id="c"),
                                            ),
                                            [],
                                        )
                                    ],
                                ),
                            ],
                        ),
                        fixture_node(
                            fixture_primitive("d2", self.d2_members), leaf=True
                        ),
                        fixture_node(
                            fixture_primitive("d3", ["pcs_rsc_order_set_1"]),
                        ),
                    ],
                ),
            ],
        )
        self.assertEqual(
            expected,
            resource.get_resource_relations_tree(
                self.env_assist.get_env(), "d1"
            ),
        )



from unittest import TestCase
from unittest.mock import patch
from socket import gaierror

from pcs.common.corosync_conf import (
    CorosyncNodeAddressDto,
    CorosyncNodeDto,
)
from pcs.lib.corosync.node import (
    CorosyncNode,
    CorosyncNodeAddress,
    get_address_type,
)


class AddrsPlain(TestCase):
    @staticmethod
    def _fixture_no_addrs():
        return CorosyncNode("node1", [], "1")

    @staticmethod
    def _fixture_one_addr():
        return CorosyncNode(
            "node1", [CorosyncNodeAddress("10.0.0.1", "1")], "1"
        )

    @staticmethod
    def _fixture_several_addrs():
        return CorosyncNode(
            "node1",
            [
                CorosyncNodeAddress("10.0.0.0", "0"),
                CorosyncNodeAddress("10.0.0.1", "1"),
                CorosyncNodeAddress("10.0.0.4", "4"),
                CorosyncNodeAddress("10.0.0.3", "3"),
            ],
            "1",
        )

    def test_no_addrs_set_get_one(self):
        node = self._fixture_no_addrs()
        self.assertEqual(None, node.addr_plain_for_link("0"))

    def test_no_addrs_set_get_all(self):
        node = self._fixture_no_addrs()
        self.assertEqual([], node.addrs_plain())

    def test_no_addrs_set_get_all_except(self):
        node = self._fixture_no_addrs()
        self.assertEqual([], node.addrs_plain(except_link="0"))

    def test_one_addr_set_get_it(self):
        node = self._fixture_one_addr()
        self.assertEqual("10.0.0.1", node.addr_plain_for_link("1"))

    def test_one_addr_set_get_another(self):
        node = self._fixture_one_addr()
        self.assertEqual(None, node.addr_plain_for_link(1))
        self.assertEqual(None, node.addr_plain_for_link("2"))

    def test_one_addr_set_get_all(self):
        node = self._fixture_one_addr()
        self.assertEqual(["10.0.0.1"], node.addrs_plain())

    def test_one_addr_set_get_all_except_it(self):
        node = self._fixture_one_addr()
        self.assertEqual([], node.addrs_plain(except_link="1"))

    def test_one_addr_set_get_all_except_another(self):
        node = self._fixture_one_addr()
        self.assertEqual(["10.0.0.1"], node.addrs_plain(except_link=1))
        self.assertEqual(["10.0.0.1"], node.addrs_plain(except_link="2"))

    def test_several_addrs_set_get_one_of_them(self):
        node = self._fixture_several_addrs()
        self.assertEqual("10.0.0.1", node.addr_plain_for_link("1"))

    def test_several_addrs_set_get_another(self):
        node = self._fixture_several_addrs()
        self.assertEqual(None, node.addr_plain_for_link(1))
        self.assertEqual(None, node.addr_plain_for_link("2"))

    def test_several_addrs_set_get_all(self):
        node = self._fixture_several_addrs()
        self.assertEqual(
            ["10.0.0.0", "10.0.0.1", "10.0.0.4", "10.0.0.3"], node.addrs_plain()
        )

    def test_several_addrs_set_get_all_except_one(self):
        node = self._fixture_several_addrs()
        self.assertEqual(
            ["10.0.0.0", "10.0.0.4", "10.0.0.3"],
            node.addrs_plain(except_link="1"),
        )

    def test_several_addrs_set_get_all_except_another(self):
        node = self._fixture_several_addrs()
        self.assertEqual(
            ["10.0.0.0", "10.0.0.1", "10.0.0.4", "10.0.0.3"],
            node.addrs_plain(except_link=1),
        )
        self.assertEqual(
            ["10.0.0.0", "10.0.0.1", "10.0.0.4", "10.0.0.3"],
            node.addrs_plain(except_link="2"),
        )


class CorosyncNodeToDto(TestCase):
    def test_no_addrs(self):
        self.assertEqual(
            CorosyncNode("node1", [], "1").to_dto(),
            CorosyncNodeDto("node1", "1", []),
        )

    def test_all_addr_types(self):
        self.assertEqual(
            CorosyncNode(
                "node1",
                [
                    CorosyncNodeAddress("10.0.0.1", "0"),
                    CorosyncNodeAddress("node1.domain", "1"),
                    CorosyncNodeAddress("fe80::5054:ff:fe81:1", "2"),
                ],
                "1",
            ).to_dto(),
            CorosyncNodeDto(
                "node1",
                "1",
                [
                    CorosyncNodeAddressDto("10.0.0.1", "0", "IPv4"),
                    CorosyncNodeAddressDto("node1.domain", "1", "FQDN"),
                    CorosyncNodeAddressDto("fe80::5054:ff:fe81:1", "2", "IPv6"),
                ],
            ),
        )


class GetAddrType(TestCase):
    def assert_call(self, expected_type, addr, resolve):
        self.assertEqual(expected_type, get_address_type(addr, resolve=resolve))

    def test_ipv4_resolve_false(self):
        self.assert_call("IPv4", "10.0.0.1", False)

    def test_ipv6_resolve_false(self):
        self.assert_call("IPv6", "fe80::5054:ff:fe81:1", False)

    def test_fqdn_resolve_false(self):
        self.assert_call("FQDN", "node1.domain", False)

    def test_ipv4_resolve_true(self):
        self.assert_call("IPv4", "10.0.0.1", True)

    def test_ipv6_resolve_true(self):
        self.assert_call("IPv6", "fe80::5054:ff:fe81:1", True)

    @patch("pcs.lib.corosync.node.socket.getaddrinfo")
    def test_fqdn_resolvable(self, mock_getaddrinfo):
        mock_getaddrinfo.return_value = None
        self.assertEqual("FQDN", get_address_type("node1.domain", resolve=True))

    @patch("pcs.lib.corosync.node.socket.getaddrinfo")
    def test_fqdn_unresolvable(self, mock_getaddrinfo):
        mock_getaddrinfo.side_effect = gaierror()
        self.assertEqual(
            "unresolvable",
            get_address_type("node1.domain", resolve=True),
        )



from typing import Any, Dict, List, Optional

from pcs.common.interface.dto import to_dict
from pcs.lib import resource_agent
from pcs.lib.commands.resource_agent import _complete_agent_list
from pcs.lib.env import LibraryEnvironment


# TODO return a list of DTOs
# for now, it is transformed to a list of dicts for backward compatibility
def list_agents(
    lib_env: LibraryEnvironment,
    describe: bool = True,
    search: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    List all stonith agents on the local host, optionally filtered and described

    describe -- load and return agents' description as well
    search -- return only agents which name contains this string
    """
    runner = lib_env.cmd_runner()
    agent_names = resource_agent.list_stonith_agents(runner)
    return _complete_agent_list(
        runner, agent_names, describe, search, resource_agent.StonithAgent
    )


# TODO return a DTO
# for now, it is transformed to a dict for backward compatibility
def describe_agent(
    lib_env: LibraryEnvironment, agent_name: str
) -> Dict[str, Any]:
    """
    Get agent's description (metadata) in a structure

    agent_name -- name of the agent (not containing "stonith:" prefix)
    """
    agent = resource_agent.find_valid_stonith_agent_by_name(
        lib_env.report_processor,
        lib_env.cmd_runner(),
        agent_name,
        absent_agent_supported=False,
    )
    return to_dict(agent.get_full_info())



import json
from textwrap import dedent
from unittest import TestCase

from pcs import settings
from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_tmp_file,
    skip_unless_root,
)
from pcs_test.tools.pcs_runner import PcsRunner


@skip_unless_root()
class SetupLocal(AssertPcsMixin, TestCase):
    def setUp(self):
        self.corosync_conf_file = get_tmp_file(
            "tier1_cluster_setup_local_corosync.conf"
        )
        self.known_hosts_file = get_tmp_file(
            "tier1_cluster_setup_local_known-hosts"
        )
        self.pcs_runner = PcsRunner(
            cib_file=None,
            corosync_conf_opt=self.corosync_conf_file.name,
            mock_settings={
                "pcsd_known_hosts_location": self.known_hosts_file.name,
            },
        )

    def tearDown(self):
        self.corosync_conf_file.close()
        if not self.known_hosts_file.closed:
            self.known_hosts_file.close()

    def fixture_known_hosts(self, node_list):
        data = {
            "format_version": 1,
            "data_version": 1,
            "known_hosts": {},
        }
        for node in node_list:
            data["known_hosts"][node["name"]] = {
                "dest_list": [{"addr": f"{node['addr']}", "port": 2224}],
                "token": f"{node['name']}_token",
            }
        self.known_hosts_file.write(json.dumps(data))
        self.known_hosts_file.flush()

    @staticmethod
    def fixture_corosync_conf_minimal(node1_addr, node2_addr):
        return dedent(
            f"""\
            totem {{
                version: 2
                cluster_name: cluster_name
                transport: knet
                crypto_cipher: aes256
                crypto_hash: sha256
            }}

            nodelist {{
                node {{
                    ring0_addr: {node1_addr}
                    name: node1
                    nodeid: 1
                }}

                node {{
                    ring0_addr: {node2_addr}
                    name: node2
                    nodeid: 2
                }}
            }}

            quorum {{
                provider: corosync_votequorum
                two_node: 1
            }}

            logging {{
                to_logfile: yes
                logfile: {settings.corosync_log_file}
                to_syslog: yes
                timestamp: on
            }}
            """
        )

    def test_file_already_exists(self):
        self.fixture_known_hosts(
            [
                {"name": "node1", "addr": "10.0.1.1"},
                {"name": "node2", "addr": "10.0.1.2"},
            ]
        )
        self.corosync_conf_file.write("some already existing content")
        self.assert_pcs_fail(
            "cluster setup cluster_name node1 node2".split(),
            dedent(
                # pylint: disable=line-too-long
                f"""\
                Error: Corosync configuration file '{self.corosync_conf_file.name}' already exists, use --overwrite to overwrite existing file(s)
                No addresses specified for host 'node1', using '10.0.1.1'
                No addresses specified for host 'node2', using '10.0.1.2'
                """
            ),
        )
        self.corosync_conf_file.seek(0)
        self.assertEqual(
            self.corosync_conf_file.read(), "some already existing content"
        )

    def test_minimal_no_known_hosts(self):
        self.known_hosts_file.close()
        self.assert_pcs_success(
            # need to use --force for not failing on unresolvable addresses
            "cluster setup cluster_name node1 node2 --force --overwrite".split(),
            dedent(
                # pylint: disable=line-too-long
                f"""\
                Warning: Unable to read the known-hosts file: No such file or directory: '{self.known_hosts_file.name}'
                No addresses specified for host 'node1', using 'node1'
                No addresses specified for host 'node2', using 'node2'
                Warning: Unable to resolve addresses: 'node1', 'node2'
                """
            ),
        )
        self.assertEqual(
            self.corosync_conf_file.read(),
            self.fixture_corosync_conf_minimal("node1", "node2"),
        )

    def test_minimal_all_known_hosts(self):
        self.fixture_known_hosts(
            [
                {"name": "node1", "addr": "10.0.1.1"},
                {"name": "node2", "addr": "10.0.1.2"},
            ]
        )
        self.assert_pcs_success(
            "cluster setup cluster_name node1 node2 --overwrite".split(),
            dedent(
                """\
                No addresses specified for host 'node1', using '10.0.1.1'
                No addresses specified for host 'node2', using '10.0.1.2'
                """
            ),
        )
        self.assertEqual(
            self.corosync_conf_file.read(),
            self.fixture_corosync_conf_minimal("10.0.1.1", "10.0.1.2"),
        )

    def test_multiple_options(self):
        self.fixture_known_hosts([])
        self.assert_pcs_success(
            (
                "cluster setup cluster_name "
                "node1 addr=127.0.0.1 addr=127.0.1.1 addr=127.0.2.3 "
                "node2 addr=127.0.0.2 addr=127.0.1.2 addr=127.0.2.2 "
                "node3 addr=127.0.0.3 addr=127.0.1.3 addr=127.0.2.1 "
                "transport knet ip_version=ipv4 link_mode=passive "
                "link linknumber=2 link_priority=100 mcastport=12345 "
                "ping_interval=1 ping_precision=2 ping_timeout=3 pong_count=4 "
                "transport=sctp "
                "link linknumber=1 transport=udp "
                "compression level=2 model=zlib threshold=10 "
                "crypto cipher=aes256 hash=sha512 model=openssl "
                "totem consensus=0 downcheck=1 token=12 "
                "quorum last_man_standing=1 last_man_standing_window=10 "
                "--overwrite"
            ).split()
        )
        self.assertEqual(
            self.corosync_conf_file.read(),
            dedent(
                """\
                totem {
                    version: 2
                    cluster_name: cluster_name
                    transport: knet
                    consensus: 0
                    downcheck: 1
                    token: 12
                    ip_version: ipv4
                    link_mode: passive
                    knet_compression_level: 2
                    knet_compression_model: zlib
                    knet_compression_threshold: 10
                    crypto_cipher: aes256
                    crypto_hash: sha512
                    crypto_model: openssl

                    interface {
                        knet_transport: udp
                        linknumber: 1
                    }

                    interface {
                        knet_link_priority: 100
                        knet_ping_interval: 1
                        knet_ping_precision: 2
                        knet_ping_timeout: 3
                        knet_pong_count: 4
                        knet_transport: sctp
                        linknumber: 2
                        mcastport: 12345
                    }
                }

                nodelist {
                    node {
                        ring0_addr: 127.0.0.1
                        ring1_addr: 127.0.1.1
                        ring2_addr: 127.0.2.3
                        name: node1
                        nodeid: 1
                    }

                    node {
                        ring0_addr: 127.0.0.2
                        ring1_addr: 127.0.1.2
                        ring2_addr: 127.0.2.2
                        name: node2
                        nodeid: 2
                    }

                    node {
                        ring0_addr: 127.0.0.3
                        ring1_addr: 127.0.1.3
                        ring2_addr: 127.0.2.1
                        name: node3
                        nodeid: 3
                    }
                }

                quorum {
                    provider: corosync_votequorum
                    last_man_standing: 1
                    last_man_standing_window: 10
                }

                logging {
                    to_logfile: yes"""
                f"""
                    logfile: {settings.corosync_log_file}"""
                """
                    to_syslog: yes
                    timestamp: on
                }
                """
            ),
        )

    def test_failure(self):
        # pylint: disable=line-too-long
        self.fixture_known_hosts([])
        self.assert_pcs_fail(
            (
                "cluster setup cluster_name "
                "node1 addr=127.0.0.1 addr=127.0.1.1.2 addr=127.0.2.3 "
                "node2 addr=127.0.0.2 addr=127.0.2.2 "
                "node3 addr=127.0.0.3 addr=127.0.1.3 addr=127.0.2.1 "
                "transport knet ip_version=ipv4 link_mode=passive "
                "link linknumber=2 link_priority=100 mcastport=123450 "
                "ping_interval=1 ping_precision=2 ping_timeout=3 pong__count=4 "
                "transport=sctp "
                "link linknumber=3 transport=udp "
                "compression level=2 model=zlib threshold=10 "
                "crypto hash=sha512 model=openssl "
                "totem consensus=0 downcheck=1 token=12 "
                "quorum lst_man_standing=1 last_man_standing_window=10"
            ).split(),
            stdout_full=dedent(
                """\
                Error: Unable to resolve addresses: '127.0.1.1.2', use --force to override
                Error: All nodes must have the same number of addresses; nodes 'node1', 'node3' have 3 addresses; node 'node2' has 2 addresses
                Error: invalid link option 'pong__count', allowed options are: 'link_priority', 'linknumber', 'mcastport', 'ping_interval', 'ping_precision', 'ping_timeout', 'pong_count', 'transport'
                Error: '123450' is not a valid mcastport value, use a port number (1..65535)
                Error: Cannot set options for non-existent link '3', existing links: '0', '1', '2'
                Error: invalid quorum option 'lst_man_standing', allowed options are: 'auto_tie_breaker', 'last_man_standing', 'last_man_standing_window', 'wait_for_all'
                Error: If quorum option 'last_man_standing_window' is enabled, quorum option 'last_man_standing' must be enabled as well
                Error: Errors have occurred, therefore pcs is unable to continue
                """
            ),
        )
        self.assertEqual(self.corosync_conf_file.read(), "")



from unittest import mock, TestCase


from pcs.common.services import errors
from pcs.common.services.drivers import SystemdDriver
from pcs.common.services.interfaces import ExecutorInterface
from pcs.common.services.types import ExecutorResult


def service_name(service, instance=None):
    if instance:
        service += f"@{instance}"
    return f"{service}.service"


class Base(TestCase):
    def setUp(self):
        self.mock_executor = mock.MagicMock(spec_set=ExecutorInterface)
        self.service = "service_name"
        self.instance = "instance_name"
        self.binary = "path"
        self.driver = SystemdDriver(self.mock_executor, self.binary, [])


class BaseTestMixin:
    subcmd = None
    exception = None
    driver_callback = staticmethod(lambda: None)

    def test_success(self):
        self.mock_executor.run.return_value = ExecutorResult(0, "", "")
        self.driver_callback(self.service)
        self.mock_executor.run.assert_called_once_with(
            [self.binary, self.subcmd, service_name(self.service)]
        )

    def test_instance_success(self):
        self.mock_executor.run.return_value = ExecutorResult(0, "", "")
        self.driver_callback(self.service, self.instance)
        self.mock_executor.run.assert_called_once_with(
            [
                self.binary,
                self.subcmd,
                service_name(self.service, self.instance),
            ]
        )

    def test_failure(self):
        result = ExecutorResult(1, "stdout", "stderr")
        self.mock_executor.run.return_value = result
        with self.assertRaises(self.exception) as cm:
            self.driver_callback(self.service)

        self.assertEqual(cm.exception.service, self.service)
        self.assertEqual(cm.exception.message, result.joined_output)
        self.assertIsNone(cm.exception.instance)
        self.mock_executor.run.assert_called_once_with(
            [self.binary, self.subcmd, service_name(self.service)]
        )

    def test_instace_failure(self):
        result = ExecutorResult(1, "stdout", "stderr")
        self.mock_executor.run.return_value = result
        with self.assertRaises(self.exception) as cm:
            self.driver_callback(self.service, self.instance)

        self.assertEqual(cm.exception.service, self.service)
        self.assertEqual(cm.exception.message, result.joined_output)
        self.assertEqual(cm.exception.instance, self.instance)
        self.mock_executor.run.assert_called_once_with(
            [
                self.binary,
                self.subcmd,
                service_name(self.service, self.instance),
            ]
        )


class StartTest(Base, BaseTestMixin):
    subcmd = "start"
    exception = errors.StartServiceError

    def setUp(self):
        super().setUp()
        self.driver_callback = self.driver.start


class StopTest(Base, BaseTestMixin):
    subcmd = "stop"
    exception = errors.StopServiceError

    def setUp(self):
        super().setUp()
        self.driver_callback = self.driver.stop


class EnableTest(Base, BaseTestMixin):
    subcmd = "enable"
    exception = errors.EnableServiceError

    def setUp(self):
        super().setUp()
        self.driver_callback = self.driver.enable


class DisableTest(Base, BaseTestMixin):
    subcmd = "disable"
    exception = errors.DisableServiceError

    def setUp(self):
        super().setUp()
        # pylint: disable=protected-access
        self.driver._available_services = [self.service]
        self.driver_callback = self.driver.disable

    def test_not_intalled(self):
        # pylint: disable=protected-access
        self.driver._available_services = [f"not_{self.service}"]
        self.driver_callback(self.service)
        self.mock_executor.run.assert_not_called()


class IsEnabledTest(Base):
    def test_enabled(self):
        self.mock_executor.run.return_value = ExecutorResult(0, "enabled", "")
        self.assertTrue(self.driver.is_enabled(self.service))
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "is-enabled", service_name(self.service)]
        )

    def test_instance_enabled(self):
        self.mock_executor.run.return_value = ExecutorResult(0, "enabled", "")
        self.assertTrue(self.driver.is_enabled(self.service, self.instance))
        self.mock_executor.run.assert_called_once_with(
            [
                self.binary,
                "is-enabled",
                service_name(self.service, self.instance),
            ]
        )

    def test_disabled(self):
        self.mock_executor.run.return_value = ExecutorResult(1, "disabled", "")
        self.assertFalse(self.driver.is_enabled(self.service))
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "is-enabled", service_name(self.service)]
        )

    def test_failure(self):
        self.mock_executor.run.return_value = ExecutorResult(1, "", "error")
        self.assertFalse(self.driver.is_enabled(self.service))
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "is-enabled", service_name(self.service)]
        )


class IsRunningTest(Base):
    def test_running(self):
        self.mock_executor.run.return_value = ExecutorResult(0, "active", "")
        self.assertTrue(self.driver.is_running(self.service))
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "is-active", service_name(self.service)]
        )

    def test_instance_running(self):
        self.mock_executor.run.return_value = ExecutorResult(0, "active", "")
        self.assertTrue(self.driver.is_running(self.service, self.instance))
        self.mock_executor.run.assert_called_once_with(
            [
                self.binary,
                "is-active",
                service_name(self.service, self.instance),
            ]
        )

    def test_not_running(self):
        self.mock_executor.run.return_value = ExecutorResult(3, "inactive", "")
        self.assertFalse(self.driver.is_running(self.service))
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "is-active", service_name(self.service)]
        )

    def test_failure(self):
        self.mock_executor.run.return_value = ExecutorResult(1, "", "error")
        self.assertFalse(self.driver.is_running(self.service))
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "is-active", service_name(self.service)]
        )


class IsInstalledTest(Base):
    def test_installed(self):
        output = (
            "service1.service        disabled\n"
            "something.target        enabled\n"
            "abc.service             enabled\n"
            "xyz.service             disabled\n"
            "serviceabcxyz.service   enabled\n"
            f"{self.service}.service disabled\n"
        )
        self.mock_executor.run.return_value = ExecutorResult(0, output, "")
        self.assertTrue(self.driver.is_installed(self.service))
        # Intetionally called twice to make sure that unit files listing is
        # done only once
        self.assertTrue(self.driver.is_installed(self.service))
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "list-unit-files", "--full"]
        )

    def test_not_installed(self):
        output = (
            "service1.service        disabled\n"
            "something.target        enabled\n"
            "abc.service             enabled\n"
            "xyz.service             disabled\n"
            "serviceabcxyz.service   enabled\n"
        )
        self.mock_executor.run.return_value = ExecutorResult(0, output, "")
        self.assertFalse(self.driver.is_installed(self.service))
        # Intetionally called twice to make sure that unit files listing is
        # done only once
        self.assertFalse(self.driver.is_installed(self.service))
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "list-unit-files", "--full"]
        )


class GetAvailableServicesTest(Base):
    def test_success(self):
        output = (
            "service1.service        disabled\n"
            "something.target        enabled\n"
            "abc.service             enabled\n"
            "xyz.service             disabled\n"
            "serviceabcxyz.service   enabled\n"
        )
        self.mock_executor.run.return_value = ExecutorResult(0, output, "")
        self.assertEqual(
            self.driver.get_available_services(),
            ["service1", "abc", "xyz", "serviceabcxyz"],
        )
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "list-unit-files", "--full"]
        )

    def test_failure(self):
        self.mock_executor.run.return_value = ExecutorResult(1, "", "error")
        self.assertEqual(self.driver.get_available_services(), [])
        self.mock_executor.run.assert_called_once_with(
            [self.binary, "list-unit-files", "--full"]
        )



# pylint: disable=protected-access
from unittest import TestCase

from pcs.common import str_tools as tools


class JoinMultilinesTest(TestCase):
    def test_empty_input(self):
        self.assertEqual("", tools.join_multilines([]))

    def test_two_strings(self):
        self.assertEqual("a\nb", tools.join_multilines(["a", "b"]))

    def test_strip(self):
        self.assertEqual("a\nb", tools.join_multilines(["  a\n", "  b\n"]))

    def test_skip_empty(self):
        self.assertEqual(
            "a\nb", tools.join_multilines(["  a\n", "   \n", "  b\n"])
        )

    def test_multiline(self):
        self.assertEqual(
            "a\nA\nb\nB", tools.join_multilines(["a\nA\n", "b\nB\n"])
        )


class IndentTest(TestCase):
    def test_indent_list_of_lines(self):
        self.assertEqual(
            tools.indent(["first", "second"]), ["  first", "  second"]
        )


class FormatOptionalTest(TestCase):
    def test_info_key_is_falsy(self):
        self.assertEqual("", tools.format_optional("", "{0}: "))

    def test_info_key_is_not_falsy(self):
        self.assertEqual("A: ", tools.format_optional("A", "{0}: "))

    def test_default_value(self):
        self.assertEqual(
            "DEFAULT", tools.format_optional("", "{0}: ", "DEFAULT")
        )

    def test_integer_zero_is_not_falsy(self):
        self.assertEqual("0: ", tools.format_optional(0, "{0}: "))


class IsMultipleTest(TestCase):
    def test_unsupported(self):
        def empty_func():
            pass

        self.assertFalse(tools._is_multiple(empty_func()))

    def test_empty_string(self):
        self.assertFalse(tools._is_multiple(""))

    def test_string(self):
        self.assertFalse(tools._is_multiple("some string"))

    def test_list_empty(self):
        self.assertTrue(tools._is_multiple(list()))

    def test_list_single(self):
        self.assertFalse(tools._is_multiple(["the only list item"]))

    def test_list_multiple(self):
        self.assertTrue(tools._is_multiple(["item1", "item2"]))

    def test_dict_empty(self):
        self.assertTrue(tools._is_multiple(dict()))

    def test_dict_single(self):
        self.assertFalse(tools._is_multiple({"the only index": "something"}))

    def test_dict_multiple(self):
        self.assertTrue(tools._is_multiple({1: "item1", 2: "item2"}))

    def test_set_empty(self):
        self.assertTrue(tools._is_multiple(set()))

    def test_set_single(self):
        self.assertFalse(tools._is_multiple({"the only set item"}))

    def test_set_multiple(self):
        self.assertTrue(tools._is_multiple({"item1", "item2"}))

    def test_integer_zero(self):
        self.assertTrue(tools._is_multiple(0))

    def test_integer_one(self):
        self.assertFalse(tools._is_multiple(1))

    def test_integer_negative_one(self):
        self.assertFalse(tools._is_multiple(-1))

    def test_integer_more(self):
        self.assertTrue(tools._is_multiple(3))


class AddSTest(TestCase):
    def test_add_s(self):
        self.assertEqual(tools._add_s("fedora"), "fedoras")

    def test_add_es_s(self):
        self.assertEqual(tools._add_s("bus"), "buses")

    def test_add_es_x(self):
        self.assertEqual(tools._add_s("box"), "boxes")

    def test_add_es_o(self):
        self.assertEqual(tools._add_s("zero"), "zeroes")

    def test_add_es_ss(self):
        self.assertEqual(tools._add_s("address"), "addresses")

    def test_add_es_sh(self):
        self.assertEqual(tools._add_s("wish"), "wishes")

    def test_add_es_ch(self):
        self.assertEqual(tools._add_s("church"), "churches")


class GetPluralTest(TestCase):
    def test_common_plural(self):
        self.assertEqual("are", tools.get_plural("is"))

    def test_add_s(self):
        self.assertEqual("pieces", tools.get_plural("piece"))


class FormatPluralTest(TestCase):
    def test_is_sg(self):
        self.assertEqual("is", tools.format_plural(1, "is"))

    def test_is_pl(self):
        self.assertEqual("are", tools.format_plural(2, "is"))

    def test_do_sg(self):
        self.assertEqual("does", tools.format_plural("he", "does"))

    def test_do_pl(self):
        self.assertEqual("do", tools.format_plural(["he", "she"], "does"))

    def test_have_sg(self):
        self.assertEqual("has", tools.format_plural("he", "has"))

    def test_have_pl(self):
        self.assertEqual("have", tools.format_plural(["he", "she"], "has"))

    def test_plural_sg(self):
        self.assertEqual(
            "singular", tools.format_plural(1, "singular", "plural")
        )

    def test_plural_pl(self):
        self.assertEqual(
            "plural", tools.format_plural(10, "singular", "plural")
        )

    def test_regular_sg(self):
        self.assertEqual("greeting", tools.format_plural(1, "greeting"))

    def test_regular_pl(self):
        self.assertEqual("greetings", tools.format_plural(10, "greeting"))


class FormatList(TestCase):
    def test_empty_list(self):
        self.assertEqual(tools.format_list([]), "")

    def test_one_item(self):
        self.assertEqual(tools.format_list(["item"]), "'item'")

    def test_multiple_items(self):
        self.assertEqual(
            tools.format_list(["item2", "item0", "item1"]),
            "'item0', 'item1', 'item2'",
        )

    def test_custom_separator(self):
        self.assertEqual(
            tools.format_list(["item2", "item0", "item1"], separator=" and "),
            "'item0' and 'item1' and 'item2'",
        )


class FormatListCustomLastSeparatort(TestCase):
    def test_empty_list(self):
        self.assertEqual(
            tools.format_list_custom_last_separator([], " and "), ""
        )

    def test_one_item(self):
        self.assertEqual(
            tools.format_list_custom_last_separator(["item"], " and "), "'item'"
        )

    def test_two_items(self):
        self.assertEqual(
            tools.format_list_custom_last_separator(
                ["item1", "item2"], " and "
            ),
            "'item1' and 'item2'",
        )

    def test_multiple_items(self):
        self.assertEqual(
            tools.format_list_custom_last_separator(
                ["item2", "item0", "item1", "item3"], " and "
            ),
            "'item0', 'item1', 'item2' and 'item3'",
        )

    def test_custom_separator(self):
        self.assertEqual(
            tools.format_list_custom_last_separator(
                ["item2", "item0", "item1", "item3"], " or ", separator=" and "
            ),
            "'item0' and 'item1' and 'item2' or 'item3'",
        )


class FormatNameValueList(TestCase):
    def test_empty(self):
        self.assertEqual([], tools.format_name_value_list([]))

    def test_many(self):
        self.assertEqual(
            ["name1=value1", '"name=2"="value 2"', '"name 3"="value=3"'],
            tools.format_name_value_list(
                [
                    ("name1", "value1"),
                    ("name=2", "value 2"),
                    ("name 3", "value=3"),
                ]
            ),
        )


class Quote(TestCase):
    def test_no_quote(self):
        self.assertEqual("string", tools.quote("string", " "))
        self.assertEqual("string", tools.quote("string", " ="))

    def test_quote(self):
        self.assertEqual('"str ing"', tools.quote("str ing", " ="))
        self.assertEqual('"str=ing"', tools.quote("str=ing", " ="))

    def test_alternative_quote(self):
        self.assertEqual("""'st"r i"ng'""", tools.quote('st"r i"ng', " "))

    def test_escape(self):
        self.assertEqual('''"st\\"r i'ng"''', tools.quote("st\"r i'ng", " "))


class Transform(TestCase):
    def test_transform(self):
        self.assertEqual(
            tools.transform(
                ["A", "0", "C", "x", "A", "B", "a"], {"A": "a", "C": "1"}
            ),
            ["a", "0", "1", "x", "a", "B", "a"],
        )



from pcs_test.tier1.cib_resource.common import ResourceTest
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.misc import (
    get_test_resource as rc,
    outdent,
    ParametrizedTestMetaClass,
    write_data_to_tmpfile,
)


ERRORS_HAVE_OCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)


def fixture_nolive_add_report(node_name):
    # pylint: disable=line-too-long
    return outdent(
        f"""\
        Unable to check if there is a conflict with nodes set in corosync because the command does not run on a live cluster (e.g. -f was used)
        Distribution of 'pacemaker authkey' to '{node_name}' was skipped because the command does not run on a live cluster (e.g. -f was used). Please, distribute the file(s) manually.
        Running action(s) 'pacemaker_remote enable', 'pacemaker_remote start' on '{node_name}' was skipped because the command does not run on a live cluster (e.g. -f was used). Please, run the action(s) manually.
        """
    )


def fixture_nolive_remove_report(host_list):
    # pylint: disable=line-too-long
    return outdent(
        """\
        Running action(s) 'pacemaker_remote disable', 'pacemaker_remote stop' on {hosts} was skipped because the command does not run on a live cluster (e.g. -f was used). Please, run the action(s) manually.
        Removing 'pacemaker authkey' from {hosts} was skipped because the command does not run on a live cluster (e.g. -f was used). Please, remove the file(s) manually.
        """
    ).format(hosts=", ".join("'{0}'".format(host) for host in host_list))


class RemoteTest(ResourceTest):
    corosync_conf = rc("corosync.conf")

    def setUp(self):
        super().setUp()
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.pcs_runner.mock_settings = get_mock_settings("crm_resource_binary")


class NodeAddRemote(RemoteTest):
    def test_fail_on_duplicit_address_specification(self):
        self.assert_pcs_fail(
            "cluster node add-remote remote-node ADDRESS server=DIFFERENT".split(),
            "Error: invalid resource option 'server', allowed options"
            " are: 'port', 'reconnect_interval', 'trace_file', 'trace_ra'\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_on_unknown_instance_attribute_not_offer_server(self):
        self.assert_pcs_fail(
            "cluster node add-remote remote-node ADDRESS abcd=efgh".split(),
            "Error: invalid resource option 'abcd', allowed options"
            " are: 'port', 'reconnect_interval', 'trace_file', 'trace_ra', "
            "use --force to override\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_on_bad_commandline_usage(self):
        self.assert_pcs_fail(
            "cluster node add-remote".split(),
            stdout_start="\nUsage: pcs cluster node add-remote...",
        )

    def test_success(self):
        self.assert_effect(
            "cluster node add-remote node-name node-addr".split(),
            """<resources>
                <primitive class="ocf" id="node-name" provider="pacemaker"
                    type="remote"
                >
                    <instance_attributes id="node-name-instance_attributes">
                        <nvpair id="node-name-instance_attributes-server"
                            name="server" value="node-addr"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="node-name-migrate_from-interval-0s"
                            interval="0s" name="migrate_from" timeout="60s"
                        />
                        <op id="node-name-migrate_to-interval-0s" interval="0s"
                            name="migrate_to" timeout="60s"
                        />
                        <op id="node-name-monitor-interval-60s" interval="60s"
                            name="monitor" timeout="30s"
                        />
                        <op id="node-name-reload-interval-0s" interval="0s"
                            name="reload" timeout="60s"
                        />
                        <op id="node-name-start-interval-0s" interval="0s"
                            name="start" timeout="60s"
                        />
                        <op id="node-name-stop-interval-0s" interval="0s"
                            name="stop" timeout="60s"
                        />
                    </operations>
                </primitive>
            </resources>""",
            output=fixture_nolive_add_report("node-name"),
        )

    def test_success_no_default_ops(self):
        self.assert_effect(
            "cluster node add-remote node-name node-addr --no-default-ops".split(),
            """<resources>
                <primitive class="ocf" id="node-name" provider="pacemaker"
                    type="remote"
                >
                    <instance_attributes id="node-name-instance_attributes">
                        <nvpair id="node-name-instance_attributes-server"
                            name="server" value="node-addr"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="node-name-monitor-interval-60s" interval="60s"
                            name="monitor" timeout="30s"
                        />
                    </operations>
                </primitive>
            </resources>""",
            output=fixture_nolive_add_report("node-name"),
        )

    def test_fail_when_server_already_used(self):
        self.assert_effect(
            "cluster node add-remote A node-addr --no-default-ops".split(),
            """<resources>
                <primitive class="ocf" id="A" provider="pacemaker"
                    type="remote"
                >
                    <instance_attributes id="A-instance_attributes">
                        <nvpair id="A-instance_attributes-server" name="server"
                            value="node-addr"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="A-monitor-interval-60s" interval="60s"
                            name="monitor" timeout="30s"
                        />
                    </operations>
                </primitive>
            </resources>""",
            output=fixture_nolive_add_report("A"),
        )
        self.assert_pcs_fail(
            "cluster node add-remote B node-addr".split(),
            "Error: 'node-addr' already exists\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_when_server_already_used_as_guest(self):
        self.pcs_runner.corosync_conf_opt = None
        self.assert_pcs_success(
            "resource create G ocf:heartbeat:Dummy --no-default-ops".split(),
        )
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.assert_pcs_success(
            "cluster node add-guest node-name G remote-addr=node-addr".split(),
            fixture_nolive_add_report("node-name"),
        )
        self.assert_pcs_fail(
            "cluster node add-remote B node-addr".split(),
            "Error: 'node-addr' already exists\n"
            "Error: Errors have occurred, therefore pcs is unable to "
            "continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )


class NodeAddGuest(RemoteTest):
    def create_resource(self):
        self.pcs_runner.corosync_conf_opt = None
        self.assert_effect(
            "resource create G ocf:heartbeat:Dummy --no-default-ops".split(),
            """<resources>
                <primitive class="ocf" id="G" provider="heartbeat" type="Dummy">
                    <operations>
                        <op id="G-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )
        self.pcs_runner.corosync_conf_opt = self.corosync_conf

    def test_fail_on_bad_commandline_usage(self):
        self.assert_pcs_fail(
            "cluster node add-guest".split(),
            stdout_start="\nUsage: pcs cluster node add-guest...",
        )

    def test_fail_when_option_remote_node_specified(self):
        self.create_resource()
        self.assert_pcs_fail(
            "cluster node add-guest node-name G remote-node=node-name".split(),
            stdout_start="Error: invalid guest option 'remote-node',"
            " allowed options are: 'remote-addr', 'remote-connect-timeout',"
            " 'remote-port'\n",
        )

    def test_fail_when_resource_has_already_remote_node_meta(self):
        self.pcs_runner.corosync_conf_opt = None
        self.assert_pcs_success(
            (
                "resource create already-guest-node ocf:heartbeat:Dummy "
                "meta remote-node=some --force"
            ).split(),
            "Warning: this command is not sufficient for creating a guest node,"
            " use 'pcs cluster node add-guest'\n",
        )
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.assert_pcs_fail(
            "cluster node add-guest node-name already-guest-node remote-addr=a".split(),
            "Error: the resource 'already-guest-node' is already a guest node\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_on_combined_reasons(self):
        self.assert_pcs_fail(
            "cluster node add-guest node-name G a=b remote-addr=a".split(),
            "Error: invalid guest option 'a', allowed options are:"
            " 'remote-addr', 'remote-connect-timeout', 'remote-port'\n"
            "Error: resource 'G' does not exist\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_when_disallowed_option_appear(self):
        self.create_resource()
        self.assert_pcs_fail(
            "cluster node add-guest node-name G a=b remote-addr=a".split(),
            "Error: invalid guest option 'a', allowed options are:"
            " 'remote-addr', 'remote-connect-timeout', 'remote-port'\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_when_invalid_interval_appear(self):
        self.create_resource()
        self.assert_pcs_fail(
            (
                "cluster node add-guest node-name G remote-connect-timeout=A "
                "remote-addr=a"
            ).split(),
            "Error: 'A' is not a valid remote-connect-timeout value, use time"
            " interval (e.g. 1, 2s, 3m, 4h, ...)\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_when_invalid_port_appear(self):
        self.create_resource()
        self.assert_pcs_fail(
            (
                "cluster node add-guest node-name G remote-port=70000 "
                "remote-addr=a"
            ).split(),
            "Error: '70000' is not a valid remote-port value, use a port number"
            " (1..65535)\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_when_guest_node_conflicts_with_existing_id(self):
        self.create_resource()
        self.pcs_runner.corosync_conf_opt = None
        self.assert_pcs_success(
            "resource create CONFLICT ocf:heartbeat:Dummy".split()
        )
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.assert_pcs_fail(
            "cluster node add-guest CONFLICT G remote-addr=a".split(),
            "Error: 'CONFLICT' already exists\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_when_guest_node_conflicts_with_existing_guest(self):
        self.create_resource()
        self.pcs_runner.corosync_conf_opt = None
        self.assert_pcs_success("resource create H ocf:heartbeat:Dummy".split())
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.assert_pcs_success(
            "cluster node add-guest node-name G remote-addr=a".split(),
            fixture_nolive_add_report("node-name"),
        )
        self.assert_pcs_fail(
            "cluster node add-guest node-name H remote-addr=b".split(),
            "Error: 'node-name' already exists\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_when_guest_node_conflicts_with_existing_remote(self):
        self.create_resource()
        self.pcs_runner.corosync_conf_opt = None
        self.assert_pcs_success(
            "resource create R ocf:pacemaker:remote server=node-addr --force".split(),
            "Warning: this command is not sufficient for creating a remote"
            " connection, use 'pcs cluster node add-remote'\n",
        )
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.assert_pcs_fail(
            "cluster node add-guest node-name G remote-addr=node-addr".split(),
            "Error: 'node-addr' already exists\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_fail_when_guest_node_name_conflicts_with_existing_remote(self):
        self.create_resource()
        self.pcs_runner.corosync_conf_opt = None
        self.assert_pcs_success(
            "resource create R ocf:pacemaker:remote server=node-addr --force".split(),
            "Warning: this command is not sufficient for creating a remote"
            " connection, use 'pcs cluster node add-remote'\n",
        )
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.assert_pcs_fail(
            "cluster node add-guest R G remote-addr=a".split(),
            "Error: 'R' already exists\n"
            "Error: Errors have occurred, therefore pcs is unable to continue\n"
            "Unable to check if there is a conflict with nodes set in corosync "
            "because the command does not run on a live cluster (e.g. -f "
            "was used)\n",
        )

    def test_success(self):
        self.create_resource()
        self.assert_effect(
            "cluster node add-guest node-name G remote-addr=node-addr".split(),
            """<resources>
                <primitive class="ocf" id="G" provider="heartbeat" type="Dummy">
                    <meta_attributes id="G-meta_attributes">
                        <nvpair id="G-meta_attributes-remote-addr"
                            name="remote-addr" value="node-addr"
                        />
                        <nvpair id="G-meta_attributes-remote-node"
                            name="remote-node" value="node-name"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="G-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>""",
            output=fixture_nolive_add_report("node-name"),
        )

    def test_success_when_guest_node_matches_with_existing_guest(self):
        # This test belongs to pcs_test/tier0/test_resource.py as it tests
        # "resource update". But due to some fixtures it is more practical to
        # keep it here.
        self.create_resource()
        self.assert_pcs_success(
            "cluster node add-guest node-name G remote-addr=a".split(),
            fixture_nolive_add_report("node-name"),
        )
        self.pcs_runner.corosync_conf_opt = None
        self.assert_pcs_success(
            "resource update G meta remote-node=node-name".split()
        )

    def test_success_with_options(self):
        self.create_resource()
        self.assert_effect(
            (
                "cluster node add-guest node-name G remote-port=3121 "
                "remote-addr=node-addr remote-connect-timeout=80s"
            ).split(),
            """<resources>
                <primitive class="ocf" id="G" provider="heartbeat" type="Dummy">
                    <meta_attributes id="G-meta_attributes">
                        <nvpair id="G-meta_attributes-remote-addr"
                            name="remote-addr" value="node-addr"
                        />
                        <nvpair id="G-meta_attributes-remote-connect-timeout"
                            name="remote-connect-timeout" value="80s"
                        />
                        <nvpair id="G-meta_attributes-remote-node"
                            name="remote-node" value="node-name"
                        />
                        <nvpair id="G-meta_attributes-remote-port"
                            name="remote-port" value="3121"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="G-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>""",
            output=fixture_nolive_add_report("node-name"),
        )


class NodeDeleteRemoveRemote(RemoteTest):
    command = None

    def fixture_remote_node(self):
        self.pcs_runner.corosync_conf_opt = None
        self.assert_effect(
            (
                "resource create NODE-NAME ocf:pacemaker:remote "
                "server=NODE-HOST --no-default-ops --force"
            ).split(),
            """<resources>
                <primitive class="ocf" id="NODE-NAME" provider="pacemaker"
                    type="remote"
                >
                    <instance_attributes id="NODE-NAME-instance_attributes">
                        <nvpair id="NODE-NAME-instance_attributes-server"
                            name="server" value="NODE-HOST"
                        />
                    </instance_attributes>
                    <operations>
                        <op id="NODE-NAME-monitor-interval-60s" interval="60s"
                            name="monitor" timeout="30s"
                        />
                    </operations>
                </primitive>
            </resources>""",
            "Warning: this command is not sufficient for creating a remote"
            " connection, use 'pcs cluster node add-remote'\n",
        )
        self.pcs_runner.corosync_conf_opt = self.corosync_conf

    def fixture_multiple_remote_nodes(self):
        # bypass pcs validation mechanisms (including expected future
        # validation)
        write_data_to_tmpfile(
            """
            <cib epoch="557" num_updates="122" admin_epoch="0"
                validate-with="pacemaker-1.2" crm_feature_set="3.0.6"
                update-origin="rh7-3" update-client="crmd"
                cib-last-written="Thu Aug 23 16:49:17 2012"
                have-quorum="0" dc-uuid="2"
            >
              <configuration>
                <crm_config/>
                <nodes>
                </nodes>
                <resources>
                    <primitive class="ocf" id="NODE-NAME"
                        provider="pacemaker" type="remote"
                    >
                        <instance_attributes id="ia1">
                            <nvpair id="nvp1" name="server" value="HOST-A"/>
                        </instance_attributes>
                    </primitive>
                    <primitive class="ocf" id="HOST-A"
                        provider="pacemaker" type="remote"
                    >
                        <instance_attributes id="ia2">
                            <nvpair id="nvp2" name="server" value="HOST-B"/>
                        </instance_attributes>
                    </primitive>
                </resources>
                <constraints/>
              </configuration>
              <status/>
            </cib>
            """,
            self.temp_cib,
        )

    def _test_usage(self):
        self.assert_pcs_fail(
            ["cluster", "node", self.command],
            stdout_start=f"\nUsage: pcs cluster node {self.command}...",
        )

    def _test_fail_when_node_does_not_exists(self):
        self.assert_pcs_fail(
            ["cluster", "node", self.command, "not-existent"],
            "Error: remote node 'not-existent' does not appear to exist in"
            " configuration\n",
        )

    def _test_success_remove_by_host(self):
        self.fixture_remote_node()
        self.assert_effect(
            ["cluster", "node", self.command, "NODE-HOST"],
            "<resources/>",
            fixture_nolive_remove_report(["NODE-NAME"])
            + outdent(
                """\
                Deleting Resource - NODE-NAME
                """
            ),
        )

    def _test_success_remove_by_node_name(self):
        self.fixture_remote_node()
        self.assert_effect(
            ["cluster", "node", self.command, "NODE-NAME"],
            "<resources/>",
            fixture_nolive_remove_report(["NODE-NAME"])
            + outdent(
                """\
                Deleting Resource - NODE-NAME
                """
            ),
        )

    def _test_refuse_on_duplicit(self):
        self.fixture_multiple_remote_nodes()
        self.assert_pcs_fail(
            ["cluster", "node", self.command, "HOST-A"],
            "Error: more than one resource for 'HOST-A' found: "
            "'HOST-A', 'NODE-NAME', use --force to override\n"
            + ERRORS_HAVE_OCURRED,
        )

    def _test_success_remove_multiple_nodes(self):
        self.fixture_multiple_remote_nodes()
        self.assert_effect(
            ["cluster", "node", self.command, "HOST-A", "--force"],
            "<resources/>",
            (
                "Warning: more than one resource for 'HOST-A' found: "
                "'HOST-A', 'NODE-NAME'\n"
            )
            + fixture_nolive_remove_report(["HOST-A", "NODE-NAME"])
            + outdent(
                """\
                Deleting Resource - NODE-NAME
                Deleting Resource - HOST-A
                """
            ),
        )


class NodeDeleteRemote(
    NodeDeleteRemoveRemote, metaclass=ParametrizedTestMetaClass
):
    command = "delete-remote"


class NodeRemoveRemote(
    NodeDeleteRemoveRemote, metaclass=ParametrizedTestMetaClass
):
    command = "remove-remote"


class NodeDeleteRemoveGuest(RemoteTest):
    command = None

    def fixture_guest_node(self):
        self.pcs_runner.corosync_conf_opt = None
        self.assert_effect(
            (
                "resource create NODE-ID ocf:heartbeat:Dummy --no-default-ops "
                "meta remote-node=NODE-NAME remote-addr=NODE-HOST --force"
            ).split(),
            """<resources>
                <primitive class="ocf" id="NODE-ID" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="NODE-ID-meta_attributes">
                        <nvpair id="NODE-ID-meta_attributes-remote-addr"
                            name="remote-addr" value="NODE-HOST"
                        />
                        <nvpair id="NODE-ID-meta_attributes-remote-node"
                            name="remote-node" value="NODE-NAME"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="NODE-ID-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>""",
            "Warning: this command is not sufficient for creating a guest node"
            ", use 'pcs cluster node add-guest'\n",
        )
        self.pcs_runner.corosync_conf_opt = self.corosync_conf

    def _test_usage(self):
        self.assert_pcs_fail(
            ["cluster", "node", self.command],
            stdout_start=f"\nUsage: pcs cluster node {self.command}...",
        )

    def _test_fail_when_node_does_not_exists(self):
        self.assert_pcs_fail(
            ["cluster", "node", self.command, "not-existent", "--force"],
            "Error: guest node 'not-existent' does not appear to exist in"
            " configuration\n",
        )

    def assert_remove_by_identifier(self, identifier):
        self.fixture_guest_node()
        self.assert_effect(
            ["cluster", "node", self.command, identifier],
            """<resources>
                <primitive class="ocf" id="NODE-ID" provider="heartbeat"
                    type="Dummy"
                >
                    <meta_attributes id="NODE-ID-meta_attributes" />
                    <operations>
                        <op id="NODE-ID-monitor-interval-10s" interval="10s"
                            name="monitor" timeout="20s"
                        />
                    </operations>
                </primitive>
            </resources>""",
            fixture_nolive_remove_report(["NODE-NAME"]),
        )

    def _test_success_remove_by_node_name(self):
        self.assert_remove_by_identifier("NODE-NAME")

    def _test_success_remove_by_resource_id(self):
        self.assert_remove_by_identifier("NODE-ID")

    def _test_success_remove_by_resource_host(self):
        self.assert_remove_by_identifier("NODE-HOST")


class NodeDeleteGuest(
    NodeDeleteRemoveGuest, metaclass=ParametrizedTestMetaClass
):
    command = "delete-guest"


class NodeRemoveGuest(
    NodeDeleteRemoveGuest, metaclass=ParametrizedTestMetaClass
):
    command = "remove-guest"



from dataclasses import dataclass
from typing import Sequence

from pcs.common.types import DrRole
from pcs.common.interface.dto import DataTransferObject


@dataclass(frozen=True)
class DrConfigNodeDto(DataTransferObject):
    name: str


@dataclass(frozen=True)
class DrConfigSiteDto(DataTransferObject):
    site_role: DrRole
    node_list: Sequence[DrConfigNodeDto]


@dataclass(frozen=True)
class DrConfigDto(DataTransferObject):
    local_site: DrConfigSiteDto
    remote_site_list: Sequence[DrConfigSiteDto]


@dataclass(frozen=True)
class DrSiteStatusDto(DataTransferObject):
    local_site: bool
    site_role: DrRole
    status_plaintext: str
    status_successfully_obtained: bool



from typing import (
    Any,
    Callable,
    List,
    Mapping,
    Optional,
)

from pcs import utils
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers

CliCmdInterface = Callable[[Any, List[str], InputModifiers], None]


def create_router(
    cmd_map: Mapping[str, CliCmdInterface],
    usage_sub_cmd: List[str],
    default_cmd: Optional[str] = None,
) -> CliCmdInterface:
    def _router(lib: Any, argv: List[str], modifiers: InputModifiers) -> None:
        if argv:
            sub_cmd, *argv_next = argv
        else:
            if default_cmd is None:
                raise CmdLineInputError()
            sub_cmd, argv_next = default_cmd, []

        try:
            if sub_cmd not in cmd_map:
                sub_cmd = ""
                raise CmdLineInputError()
            return cmd_map[sub_cmd](lib, argv_next, modifiers)
        except CmdLineInputError as e:
            if not usage_sub_cmd:
                raise
            return utils.exit_on_cmdline_input_errror(
                e, usage_sub_cmd[0], (usage_sub_cmd[1:] + [sub_cmd])
            )

    return _router



from . import (
    codes,
    const,
    item,
    messages,
    types,
)
from .item import (
    get_severity,
    ReportItem,
    ReportItemList,
    ReportItemSeverity,
    ReportItemContext,
)
from .dto import ReportItemDto
from .processor import (
    has_errors,
    ReportProcessor,
)
from .conversions import report_dto_to_item



from pcs_test.tools.misc import outdent
from pcs_test.tools.command_env.mock_runner import Call as RunnerCall


class CorosyncShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def version(
        self,
        name="runner.corosync.version",
        version="2.4.0",
        instead=None,
        before=None,
    ):
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync", "-v"],
                stdout=outdent(
                    """\
                    Corosync Cluster Engine, version '{0}'
                    Copyright...
                    """.format(
                        version
                    )
                ),
            ),
            before=before,
            instead=instead,
        )

    def qdevice_generate_cert(
        self,
        cluster_name,
        cert_req_path="cert_path",
        stdout=None,
        stderr="",
        returncode=0,
        name="runner.corosync.qdevice_generate_cert",
    ):
        if stdout is not None and cert_req_path is not None:
            raise AssertionError(
                "Cannot specify both 'cert_req_path' and 'stdout'"
            )
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync-qdevice-net-certutil", "-r", "-n", cluster_name],
                stdout=(
                    stdout
                    if stdout is not None
                    else f"Certificate request stored in {cert_req_path}\n"
                ),
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def qdevice_get_pk12(
        self,
        cert_path="cert path",
        output_path="output_path",
        stdout=None,
        stderr="",
        returncode=0,
        name="runner.corosync.qdevice_get_pk12",
    ):
        if stdout is not None and output_path is not None:
            raise AssertionError(
                "Cannot specify both 'output_path' and 'stdout'"
            )
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync-qdevice-net-certutil", "-M", "-c", cert_path],
                stdout=(
                    stdout
                    if stdout is not None
                    else f"Certificate stored in {output_path}\n"
                ),
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def quorum_status(
        self,
        node_list=None,
        stdout=None,
        stderr="",
        returncode=0,
        name="runner.corosync.quorum_status",
    ):
        if bool(node_list) == bool(stdout):
            raise AssertionError(
                "Exactly one of 'node_list', 'stdout' must be specified"
            )
        if node_list:
            stdout = outdent(
                """\
            Quorum information
            ------------------
            Date:             Fri Jan 16 13:03:28 2015
            Quorum provider:  corosync_votequorum
            Nodes:            {nodes_num}
            Node ID:          1
            Ring ID:          19860
            Quorate:          Yes\n
            Votequorum information
            ----------------------
            Expected votes:   {nodes_num}
            Highest expected: {nodes_num}
            Total votes:      {nodes_num}
            Quorum:           {quorum_num}
            Flags:            Quorate\n
            Membership information
            ----------------------
                Nodeid      Votes    Qdevice Name
            {nodes}\
            """
            ).format(
                nodes_num=len(node_list),
                quorum_num=(len(node_list) // 2) + 1,
                nodes="".join(
                    [
                        _quorum_status_node_fixture(node_id, node)
                        for node_id, node in enumerate(node_list, 1)
                    ]
                ),
            )
        self.__calls.place(
            name,
            RunnerCall(
                ["corosync-quorumtool", "-p"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )


def _quorum_status_node_fixture(node_id, node_name, votes=1, is_local=False):
    _local = " (local)" if is_local else ""
    return (
        f"         {node_id}          {votes}         NR {node_name}{_local}\n"
    )



from textwrap import dedent

from pcs_test.tools.misc import outdent


QDEVICE_HOST = "qdevice.host"
CLUSTER_NAME = "myCluster"


def get_two_node(nodes_num):
    if nodes_num == 2:
        return [("two_node", "1")]
    return []


def _corosync_options_fixture(option_list, indent_level=2):
    indent = indent_level * 4 * " "
    return "".join(
        [f"{indent}{option}: {value}\n" for option, value in option_list]
    )


def corosync_conf_fixture(node_list=(), quorum_options=(), qdevice_net=False):
    nodes = []
    for node in node_list:
        nodes.append(
            dedent(
                """\
                node {{
            {options}    }}
            """
            ).format(options=_corosync_options_fixture(node))
        )
    device = ""
    if qdevice_net:
        device = outdent(
            f"""
                device {{
                    model: net

                    net {{
                        host: {QDEVICE_HOST}
                    }}
                }}
            """
        )
    return dedent(
        """\
        totem {{
            version: 2
            cluster_name: {cluster_name}
            transport: knet
        }}

        nodelist {{
        {nodes}}}

        quorum {{
            provider: corosync_votequorum
        {quorum}{device}}}

        logging {{
            to_logfile: yes
            logfile: /var/log/cluster/corosync.log
            to_syslog: yes
        }}
        """
    ).format(
        cluster_name=CLUSTER_NAME,
        nodes="\n".join(nodes),
        quorum=_corosync_options_fixture(quorum_options, indent_level=1),
        device=device,
    )


def corosync_node_fixture(node_id, node, addrs):
    return [(f"ring{i}_addr", addr) for i, addr in enumerate(addrs)] + [
        ("name", node),
        ("nodeid", str(node_id)),
    ]


def node_fixture(node, node_id, addr_sufix=""):
    return corosync_node_fixture(node_id, node, [f"{node}{addr_sufix}"])


TOTEM_TEMPLATE = """\
totem {{
    transport: {transport_type}\
{totem_options}{transport_options}{compression_options}{crypto_options}
}}
"""


def fixture_totem(
    transport_type="knet",
    transport_options=None,
    compression_options=None,
    crypto_options=None,
    totem_options=None,
):
    def options_fixture(options, prefix=""):
        options = options or {}
        template = "\n    {prefix}{option}: {value}"
        return "".join(
            [
                template.format(prefix=prefix, option=o, value=v)
                for o, v in sorted(options.items())
            ]
        )

    return TOTEM_TEMPLATE.format(
        transport_type=transport_type,
        transport_options=options_fixture(transport_options),
        compression_options=options_fixture(
            compression_options, prefix="knet_compression_"
        ),
        crypto_options=options_fixture(crypto_options, prefix="crypto_"),
        totem_options=options_fixture(totem_options),
    )
